/***************************************************************************
 *   Copyright (C) 2010 by Naomasa Matsubayashi   *
 *   fadis@quaternion.sakura.ne.jp   *
 *                                                                         *
 *   All rights reserved.                                                  *
 *                                                                         *
 * Redistribution and use in source and binary forms, with or without      *
 * modification, are permitted provided that the following conditions are  *
 * met:                                                                    *
 *                                                                         *
 *  1. Redistributions of source code must retain the above copyright      *
 *     notice, this list of conditions and the following disclaimer.       *
 *  2. Redistributions in binary form must reproduce the above copyright   *
 *     notice, this list of conditions and the following disclaimer in the *
 *     documentation and/or other materials provided with the distribution.*
 *                                                                         *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS     *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,   *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    *
 ***************************************************************************/


#include "inc/hw_types.h"
#include "driverlib/debug.h"
#include "driverlib/sysctl.h"
#include "drivers/display96x16x1.h"

#include <k6x8/k6x8.h>
#include <k6x8/glyph.h>
#include <k6x8/map.h>
#include <k6x8/conv.h>
#include <k6x8/draw.h>

// フルスクリーンテキストレンダラを初期化する
void fstInit( FullScreenText *_fst, const char *_string ) {
  unsigned int index;
  for( index = 0; index != VIRTUAL_LINE_NUM; index++ )
    _fst->line_head[ index ] = NULL;
  _fst->buffer_head = 0;
  _fst->string_head = _string;
  _fst->scroll = 0;
  _fst->current_line = 0;
  _fst->end_line = VIRTUAL_LINE_NUM;
  _fst->mask = 0;
  for( index = 0; index != 3; index++ ) {
    _fst->line_head[ index ] = _fst->string_head;
    _fst->string_head = drwEUCJP( _fst->buffer + X_PIXEL_NUM * index, X_CHARCTER_NUM, _fst->string_head );
  }
  _fst->current_line = 3;
}

// 内部リングバッファの先頭にテキストを一行描画する
void fstRenderLine( FullScreenText *_fst ) {
  if( _fst->line_head[ _fst->current_line ] ) {
    drwEUCJP( _fst->buffer + _fst->buffer_head * X_PIXEL_NUM, X_CHARCTER_NUM, _fst->line_head[ _fst->current_line ] );
    _fst->buffer_head = ( _fst->buffer_head + 1 ) % 3;
    _fst->current_line++;
  }
  else if( *( _fst->string_head ) ) {
    _fst->line_head[ _fst->current_line ] = _fst->string_head;
    _fst->string_head =
      drwEUCJP( _fst->buffer + _fst->buffer_head * X_PIXEL_NUM, X_CHARCTER_NUM, _fst->string_head );
    _fst->buffer_head = ( _fst->buffer_head + 1 ) % 3;
    _fst->current_line++;
  }
  else {
    _fst->line_head[ _fst->current_line ] = _fst->string_head;
    unsigned int col;
    for( col = 0; col != X_PIXEL_NUM; col++ )
      _fst->buffer [ _fst->buffer_head * X_PIXEL_NUM + col ] = 0;
    _fst->buffer_head = ( _fst->buffer_head + 1 ) % 3;
    _fst->current_line++;
    _fst->end_line = _fst->current_line;
  }
}

// テキストをスクロールする　スクロールした結果内部バッファの描画済み範囲を越えていたら描画する
void fstScroll( FullScreenText *_fst, signed int _value ) {
  if( _value > 0 ) {
    if( _fst->current_line < _fst->end_line ) {
      _fst->scroll += _value;
      while( _fst->scroll >= 8 ) {
        fstRenderLine( _fst );
        _fst->scroll -= 8;
      }
    }
  }
}

// テキストが終端に達していたら非ゼロ値を返す
int fstIsEnd( FullScreenText *_fst ) {
  return _fst->current_line == _fst->end_line;
}

// フェードイン/フェードアウト用のマスク値
// もう少し容量をケチることも出来るが、描画時のパフォーマンスに影響するのでやめておく
const uint8_t fst_fade_mask[ 2 * 5 ] = {
  // ■■
  // ■■
  0xFF, 0xFF,
  // ■■
  // ■□
  0xFF, 0xAA,
  // ■□
  // □■
  0xAA, 0x55,
  // ■□
  // □□
  0xAA, 0x00,
  // □□
  // □□
  0x00, 0x00,
};

// 内部バッファとスクロール値から画面に描画するイメージを作る
void fstMakeScrolledImage( FullScreenText *_fst ) {
  unsigned int dest_line;
  for( dest_line = 0; dest_line != 2; dest_line++ ) {
    unsigned int source_line0 = ( _fst->current_line + dest_line ) % 3;
    unsigned int source_line1 = ( _fst->current_line + dest_line +  1 ) % 3;
    unsigned int col;
    for( col = 0; col != 96; col++ ) {
      _fst->scrolled_buffer[ col + dest_line * X_PIXEL_NUM ] =
        ( ( _fst->buffer[ source_line0 * X_PIXEL_NUM + col ] >> _fst->scroll ) |
        ( _fst->buffer[ source_line1 * X_PIXEL_NUM + col ] << ( 8 - _fst->scroll ) ) )
        & fst_fade_mask[ _fst->mask * 2 + ( col & 0x01 ) ];
    }
  }
}

// フェードイン/フェードアウト用マスクをセットする 以後描画される文字には選択したマスクが適用される
void fstSetMask( FullScreenText *_fst, uint8_t _level ) {
  _fst->mask = _level;
}

// 画面に描画する
void fstDisplay( FullScreenText *_fst ) {
  fstMakeScrolledImage( _fst );
  Display96x16x1ImageDraw( _fst->scrolled_buffer, 0, 0, X_PIXEL_NUM, 2 );
}
